# -*- coding: utf-8 -*-
import time
from threading import Lock
from geeker.functions.singleton import Singleton


class Descripter:
    def __init__(self, name=None):
        self.name = name

    def __set__(self, instance, value):
        instance.__dict__[self.name] = value

    def __get__(self, instance, owner):
        if not instance:
            return self
        return instance.__dict__[self.name]

    def __delete__(self, instance):
        del instance.__dict__[self.name]


class MyType(Descripter):
    expect_type = type(None)

    def __init__(self, name=None, **opts):
        if not opts.get('expect_type'):
            raise ValueError('Missing argument except_type')
        else:
            self.expect_type = opts.get('expect_type')
        super().__init__(name=name)

    def __set__(self, instance, value):
        if not isinstance(value, self.expect_type):
            raise TypeError(f'Except {self.expect_type} but got {type(value)}')
        super().__set__(instance, value)


class InvalidSystemClock(Exception):
    """
    时钟回拨异常
    """
    pass


"""
snow 算法

符号位            41bit时间戳,精确到毫秒          共10bit(数据中心ID+机器ID)    12bit序列号,每毫秒生成4096个ID
  1     00000000000000000000000000000000000000000         0000000000                   000000000000


"""

# 64位ID的划分
WORKER_ID_BITS = 5
DATA_CENTER_ID_BITS = 5
SEQUENCE_BITS = 12

# 最大取值计算
MAX_WORKER_ID = -1 ^ (-1 << WORKER_ID_BITS)  # 2**5-1 0b11111
MAX_DATA_CENTER_ID = -1 ^ (-1 << DATA_CENTER_ID_BITS)

# 移位偏移计算
WORKER_ID_SHIFT = SEQUENCE_BITS
DATA_CENTER_ID_SHIFT = SEQUENCE_BITS + WORKER_ID_BITS
TIMESTAMP_LEFT_SHIFT = SEQUENCE_BITS + WORKER_ID_BITS + DATA_CENTER_ID_BITS

# 序号循环掩码
SEQUENCE_MASK = -1 ^ (-1 << SEQUENCE_BITS)

# 开始时间截 (2020-01-01)
START_TIME_STAMP = 1577808000000


class IdGenerator(Singleton):
    """
    分布式唯一ID生成器

    """
    _lock = Lock()

    def __init__(self, data_center_id=5, worker_id=5, sequence=0):
        """
        :param data_center_id: 数据中心ID
        :param worker_id: 机器ID
        :param sequence: 序列号
        """

        if worker_id > MAX_WORKER_ID or worker_id < 0:
            raise ValueError('参数worker_id必须在0-31之间')

        if data_center_id > MAX_DATA_CENTER_ID or data_center_id < 0:
            raise ValueError('参数data_center_id值必须在0-31之间')

        self.worker_id = worker_id
        self.data_center_id = data_center_id
        self.sequence = sequence
        # 上次计算的时间戳
        self.last_timestamp = -1

    @staticmethod
    def _gen_timestamp():
        """
        生成整数时间戳
        :return:int timestamp
        """
        return int(time.time() * 1000)

    def get_id(self):
        with self._lock:
            return self._get_id()

    def _get_id(self):
        """
        获取新ID
        """

        timestamp = self._gen_timestamp()

        # 时钟回拨
        if timestamp < self.last_timestamp:
            raise InvalidSystemClock("clock is moving backwards.")

        if timestamp == self.last_timestamp:
            self.sequence = (self.sequence + 1) & SEQUENCE_MASK
            if self.sequence == 0:
                timestamp = self._til_next_millis(self.last_timestamp)
        else:
            self.sequence = 0

        self.last_timestamp = timestamp

        new_id = ((timestamp - START_TIME_STAMP) << TIMESTAMP_LEFT_SHIFT) | \
                 (self.data_center_id << DATA_CENTER_ID_SHIFT) | (self.worker_id << WORKER_ID_SHIFT) | self.sequence
        return new_id

    def _til_next_millis(self, last_timestamp):
        """
        需要等到下一毫秒
        """
        timestamp = self._gen_timestamp()
        while timestamp <= last_timestamp:
            timestamp = self._gen_timestamp()
        return timestamp



# -*- coding: utf-8 -*-
from collections import deque
import time
import math
from queue import Queue
from threading import Lock


class BaseNeedTime:
    """
    使用deque控制执行频率
    """
    __slots__ = ()
    _func = None
    _seconds = None
    _maxlen = None
    _dq = None
    lock = Lock()

    def append_time(self, *args, **kwargs):
        # 这里需要加锁,否则并发情况下无法控制
        with self.lock:
            if len(self._dq) < self._maxlen:
                self._dq.append(time.time())
            else:
                need_time = self._seconds - (time.time() - self._dq[0])
                need_time = math.ceil(need_time)
                if need_time > 0:
                    time.sleep(need_time)
                self._dq.append(time.time())
        result = self._func(*args, **kwargs)
        return result


class BaseConcurrency:
    """
    使用queue控制并发量
    """
    __slots__ = ()
    _func = None
    _dq = None

    def control_(self, *args, **kwargs):
        self._dq.put(None)
        try:
            result = self._func(*args, **kwargs)
            return result
        finally:
            self._dq.get()


class Concurrency(BaseNeedTime, BaseConcurrency):

    def __init__(self, times: int, seconds: int = None):
        """
        seconds秒执行m次,当只传入times参数,则并发量为times
        :param times: 次数
        :param seconds: 秒数
        """

        ERR_STR = 'Parameter <%s> must be <TYPE:int> and bigger than 0 !'
        if not (isinstance(times, int) and times > 0):
            raise ValueError(ERR_STR.format(times))

        if seconds is None:
            self._dq = Queue(maxsize=times)

        else:
            if not (isinstance(seconds, int) and seconds > 0):
                raise ValueError(ERR_STR.format(seconds))
            self._dq = deque(maxlen=times)
        self._seconds = seconds
        self._maxlen = times

    def __call__(self, func):

        self._func = func

        # 这里需要把函数的参数封装进args,并传递(如果是类方法,args的第0个参数是调用者的self)
        def warp_(*args, **kwargs):
            warp_.__name__ = func.__name__
            if self._seconds is None:
                return self.control_(*args, **kwargs)
            return self.append_time(*args, **kwargs)

        return warp_



# -*- coding: utf-8 -*-

from .log_config import LogBase
import os

__all__ = ['MyLog']


class MyLog(LogBase):
    """
    功能:
        将日志分日志等级记录,并自动压缩2019-11-11.info.log.gz

    参数:
        :param dir_path: 日志记录的路径,默认是当前路径下的log文件夹
        :param logger_name: logger对象的名字
        :param info_name: 保存info等级的文件名字
        :param error_name:
        :param warning_name:
        :param debug_name:
        :param interval: 压缩日志的频率,默认是7天
        :param detail: bool值,记录日志是否为详细记录
        :param debug: 是否记录debug,默认不记录
        :param info: 是否记录info,默认记录
        :param error:
        :param warning:
    实例方法:
        get_logger()-->logger

    使用举例:
        # 记录四种类型的日志
        logger = MyLog(debug=True).get_logger()
        logger.info('info')
        logger.debug('debug')
        logger.error('error')
        logger.warning('warning')

        # # # # # # # # # # # # # # # # # # # # # # # # #

        # 只记录错误日志
        logger = MyLog(info=False,warning=False).get_logger()
        logger.info('info')
        logger.debug('debug')
        logger.error('error')
        logger.warning('warning')
    注意:
        MyLog()的实例只会同时存在一个,默认记录首次创建实例的属性.
        例如:

            mylog = MyLog('./logs/logs/')
            mylog2 = MyLog()
            logger = mylog.get_logger()
            logger2 = mylog2.get_logger()
            logger.info('info')

            logger2 = MyLog('./logs/logs2/').get_logger()
            logger2.info('info2')

            以上两个logger logger2,会以logger(第一次创建实例)的属性为准,日志会存放在./logs/logs/下



    """

    def __init__(self, log_path: str = './logs/', **kwargs):
        self.type_need(log_path, str)
        if not log_path.endswith('/'):
            log_path += '/'
        if not os.path.exists(log_path):
            os.makedirs(log_path)
        super(MyLog, self).__init__(dir_path=log_path, **kwargs)

    def get_logger(self):
        return self._get_logger()

    @staticmethod
    def type_need(parm, type_):
        if not isinstance(parm, type_):
            raise TypeError(f'expect {type_},but got {type(parm)}')



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from .mylog import *
from . import schedule
from .functions import *
from .utils import *

__all__ = ['MyType', 'MyLog', 'TimeOut',
           'schedule', 'Concurrency',
           'Singleton', 'SingletonOverride', 'MyDict', 'run_time',
           "retry", "IdGenerator", 'show_memory_info'
           ]

__UpdateTime__ = '2021/06/09 11:25'
__Version__ = "2.1"
__Author__ = 'liu YaLong'

__Description__ = """
If you has issues ,please read README.MD or
fix it by yourself .
"""



# -*- coding: utf-8 -*-

from .timeout import TimeOut
from .timeslimit import Concurrency
from .singleton import Singleton, SingletonOverride
from .mydata import MyDict
from functools import wraps
import time
import threading
import os
import psutil

from colorama import init

init(autoreset=True)

__all__ = ['TimeOut', 'Concurrency', 'Singleton', 'SingletonOverride',
           'MyDict', 'run_time', 'retry',
           'show_memory_info']


def show_memory_info():
    """
   显示当前 python 程序占用的内存大小
    """
    pid = os.getpid()
    p = psutil.Process(pid)
    info = p.memory_full_info()
    memory = info.uss / 1024. / 1024
    print("所有线程:", threading.enumerate())
    print('线程总数:', threading.active_count())
    print("共占用", memory, "MB")


def retry(retry_times: int, error_type: (tuple, Exception) = Exception, sleep_time: int = 0):
    """
    捕获指定类型异常进行重试执行函数retry_times次,每次执行间隔sleep秒
    :param retry_times: 重试次数
    :param sleep_time: 重试间隔(秒)
    :param error_type: 需要捕获的错误类型
    """

    def my_wraps0(func):
        @wraps(func)
        def my_wraps1(*args, **kwargs):
            for _ in range(retry_times + 1):
                try:
                    result = func(*args, **kwargs)
                    return result
                except error_type as e:
                    if _ == retry_times:
                        raise e
                    time.sleep(sleep_time)
        return my_wraps1

    return my_wraps0


def run_time(func):
    # 此装饰器，用来调试函数运行时间及执行流程
    @wraps(func)  # 保留源信息
    def mywarps(*args, **kwargs):
        func_name = func.__name__
        start_time = time.time()
        print(f'''\033[32mSTART {func_name}{*args, kwargs}\033[0m''')
        try:
            aa = func(*args, **kwargs)
        finally:
            cost_time = time.time() - start_time
            print(f'''\033[32m{func_name}{*args, kwargs} takes <%.4f> seconds\033[0m''' % cost_time)
            print(f'''\033[32mSTOP {func_name}{*args, kwargs}\033[0m''')
        return aa

    return mywarps



# -*- coding: utf-8 -*-


class MyDict:

    def __getattr__(self, item):
        return None

    def __getitem__(self, item):
        return getattr(self, item)

    def __iter__(self):
        for k, v in self.__dict__.items():
            yield k, v

    def __len__(self):
        return len(self.__dict__)

    def __setitem__(self, key, value):
        setattr(self, key, value)

    def __str__(self):
        return f'''< {__name__}.MyDict object at {hex(id(MyDict))}>'''

    def __repr__(self):
        return '''MyDict object ,just like :
                {'a': 5,
                 'b': [1, 2],
                 'c': {1, 2, 3}
                 }
                '''

    @staticmethod
    def _type_check_(key, expect_type):
        if not isinstance(key, expect_type):
            raise TypeError(f"The type of the value <{key}> was wrong, expected type: <{expect_type}>")

    def append_key(self, key, value):
        if key not in self.__dict__:
            self.__dict__[key] = []

        else:
            self._type_check_(self.__dict__[key], list)

        self.__dict__[key].append(value)

    def add_key(self, key, value):
        if key not in self.__dict__:
            self.__dict__[key] = set()

        else:
            self._type_check_(self.__dict__[key], set)

        self.__dict__[key].add(value)

    def keys(self):
        return self.__dict__.keys()



# -*- coding: utf-8 -*-

import logging
import logging.handlers
from logging.handlers import TimedRotatingFileHandler
import gzip
import os
import time
from geeker.functions import Singleton


class GzTimedRotatingFileHandler(TimedRotatingFileHandler):
    def __init__(self, filename, when, interval, **kwargs):
        super(GzTimedRotatingFileHandler, self).__init__(filename, when, interval, **kwargs)

    @staticmethod
    def do_gzip(old_log):
        with open(old_log, 'rb') as old:
            with gzip.open(old_log.replace('.log', '', 1) + '.gz', 'wb') as comp_log:
                comp_log.writelines(old)
        os.remove(old_log)

    # overwrite
    def doRollover(self):
        if self.stream:
            self.stream.close()
            self.stream = None
        current_time = int(time.time())
        dst_now = time.localtime(current_time)[-1]
        t = self.rolloverAt - self.interval
        if self.utc:
            time_tuple = time.gmtime(t)
        else:
            time_tuple = time.localtime(t)
            dst_then = time_tuple[-1]
            if dst_now != dst_then:
                if dst_now:
                    addend = 3600
                else:
                    addend = -3600
                time_tuple = time.localtime(t + addend)
        dfn = self.baseFilename + "." + time.strftime(self.suffix, time_tuple)
        if os.path.exists(dfn):
            os.remove(dfn)
        if os.path.exists(self.baseFilename):
            os.rename(self.baseFilename, dfn)
            self.do_gzip(dfn)
        if self.backupCount > 0:
            for s in self.getFilesToDelete():
                os.remove(s)
        if not self.delay:
            self.stream = self._open()
        new_rollover_at = self.computeRollover(current_time)
        while new_rollover_at <= current_time:
            new_rollover_at = new_rollover_at + self.interval
        if (self.when == 'MIDNIGHT' or self.when.startswith('W')) and not self.utc:
            ds_att_rollover = time.localtime(new_rollover_at)[-1]
            if dst_now != ds_att_rollover:
                if not dst_now:  # DST kicks in before next rollover, so we need to deduct an hour
                    addend = -3600
                else:  # DST bows out before next rollover, so we need to add an hour
                    addend = 3600
                new_rollover_at += addend
        self.rolloverAt = new_rollover_at


class LogBase(Singleton):

    def __init__(self, dir_path='./logs/',
                 logger_name='special_log_name',
                 info_name='info.log',
                 error_name='error.log',
                 warning_name='warning.log',
                 debug_name='debug.log',
                 interval=7,
                 detail=False,
                 debug=False,
                 info=True,
                 error=True,
                 warning=True,
                 ):
        self.info_name = info_name
        self.error_name = error_name
        self.warning_name = warning_name
        self.debug_name = debug_name
        self.path = dir_path
        self.interval = interval
        self._logger = logging.getLogger(logger_name)
        self._debug = debug
        self._warning = warning
        self._error = error
        self._info = info
        self._detail = detail

    def __handler(self, log_name):
        handler = GzTimedRotatingFileHandler(self.path + log_name,
                                             when='D',
                                             interval=self.interval,
                                             backupCount=3,
                                             encoding='utf-8')
        return handler

    def __filter_message(self, handler, log_level):
        """
        过滤不同等级日志的其他信息,只保留当前日志等级的信息
        :param handler: handler
        :param log_level: 字符串
        :return: handler
        """
        if self._detail:
            formatter = logging.Formatter("%(asctime)s - %(filename)s - %(funcName)s - %(lineno)d - %(message)s",
                                          "%Y%m%d %H:%M:%S")
        else:
            formatter = logging.Formatter("%(asctime)s - %(message)s", "%Y%m%d %H:%M:%S")
        _filter = logging.Filter()

        handler.suffix = "%Y%m%d.log"
        handler.setFormatter(formatter)
        handler.setLevel(log_level)
        _filter.filter = lambda record: record.levelno == log_level
        handler.addFilter(_filter)
        return handler

    def _get_logger(self):

        # 添加此行，防止日志重复记录
        if not self._logger.handlers:
            # 设置日志等级,默认是 DEBUG
            self._logger.setLevel(logging.DEBUG)

            levels = [self._debug, self._info, self._warning, self._error]
            log_names = [self.debug_name, self.info_name, self.warning_name, self.error_name]
            levels_ = [10, 20, 30, 40]

            for i, lev in enumerate(levels):
                if lev:
                    _handler = self.__handler(log_names[i])
                    _handler = self.__filter_message(_handler, levels_[i])
                    # handler添加给日志对象
                    self._logger.addHandler(_handler)
        return self._logger



# -*- coding: utf-8 -*-

import os
import sys

if not __package__:
    path = os.path.join(os.path.dirname(__file__), os.pardir)
    sys.path.insert(0, path)
from geeker.cmdline import base_command  # noqa

if __name__ == '__main__':
    base_command()



# coding = utf-8
from setuptools import setup, find_packages
from geeker import __Version__

with open("README.md", "r", encoding='utf-8') as fh:
    long_description = fh.read()

setup(
    name='geeker',
    version=__Version__,
    description=(
        "Some useful functions !"
    ),
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='liuyalong',
    author_email='4379711@qq.com',
    maintainer='liuyalong',
    license='MIT License',
    packages=find_packages(),
    platforms=["all"],
    url='https://github.com/4379711/functools_lyl',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Operating System :: OS Independent',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Software Development :: Libraries'
    ],
    # 指定入口
    entry_points={
        # 添加命令行脚本
        'console_scripts': [
            'geeker=geeker.cmdline:execute'
        ],
    },

    install_requires=['colorama',
                      'click',
                      'psutil'
                      ]
)
# pip install wheel
# pip install twine

# python setup.py check                 检查错误
# python setup.py sdist bdist_wheel     编译一个tar.gz包,一个wheel包
# twine upload dist/*                   上传到pypi



# -*- coding: utf-8 -*-
import ctypes
import inspect
import time
import threading
from functools import wraps


# import traceback
# import sys


class MyThread(threading.Thread):
    def __init__(self, target, args=None, kwargs=None):
        super().__init__()
        self.func = target
        self.args = args
        self.kwargs = kwargs
        self.result = '<_^&^_@**@__what fuck!__@**@_^&^_>'
        self.exitcode = False
        self.exception = None
        # self.exc_traceback = None

    def _run(self):
        self.result = self.func(*self.args, **self.kwargs)

    def run(self):  # Overwrite run() method
        try:
            self._run()
        except Exception as e:
            self.exitcode = True
            self.exception = e
            # self.exc_traceback = sys.exc_info()

    @property
    def get_result(self):
        return self.result

    @staticmethod
    def _async_raise(tid, exctype):
        """raises the exception, performs cleanup if needed"""
        # tid = ctypes.c_long(tid)
        if not inspect.isclass(exctype):
            exctype = type(exctype)
        ctypes.pythonapi.PyThreadState_SetAsyncExc(tid, ctypes.py_object(exctype))
        # res = ctypes.pythonapi.PyThreadState_SetAsyncExc(tid, ctypes.py_object(exctype))
        # if res == 0:
        #     raise ValueError("invalid thread id")
        # elif res != 1:
        #     ctypes.pythonapi.PyThreadState_SetAsyncExc(tid, None)
        #     raise SystemError("PyThreadState_SetAsyncExc failed !")

    def stop_thread(self, ident=None):
        if ident:
            self._async_raise(ident, SystemExit)
        else:
            # self._async_raise(threading.get_ident(), SystemExit)
            for fuck in threading.enumerate():
                if fuck is self:
                    self._async_raise(fuck.ident, SystemExit)
                    break


class TimeOut:
    def __init__(self, limit_time=1):
        if not (isinstance(limit_time, (int, float)) and limit_time > 0):
            raise ValueError('The type of parameter <limit_time> must be int and greater than 0!')
        self.limit = int(limit_time * 10)

    def __raise_error(self, th):
        # exec_type = th.exc_traceback[0]
        # tmp_str = traceback.format_exception(th.exc_traceback[0], th.exc_traceback[1], th.exc_traceback[2])
        # str_ = ''.join(tmp_str[1:])
        #
        # th.stop_thread()
        #
        # # raise exec_type('\n'+str_)
        raise th.exception

    def __call__(self, func):
        @wraps(func)
        def warp_(*args, **kwargs):
            warp_.__name__ = func.__name__
            th = MyThread(target=func, args=args, kwargs=kwargs)
            th.daemon = True
            th.start()
            # Add 0.1 second here
            for _ in range(self.limit + 2):
                if th.exitcode:
                    self.__raise_error(th)

                is_result = th.get_result

                if is_result != '<_^&^_@**@__what fuck!__@**@_^&^_>':
                    return is_result

                if _ == self.limit:
                    # kill the thread by itself
                    th.stop_thread()
                    raise TimeoutError('Unfinished tasks within the specified time !')
                time.sleep(0.1)

        return warp_



# -*- coding: utf-8 -*-
# @Time    : 2019/11/13 10:03
# @Author  : Liu Yalong
# @File    : import_module.py
from importlib import import_module


def load_object(path):
    """Load an object given its absolute object path, and return it.
    object can be a class, function, variable or an instance.
    path ie: 'geeker.commands.abc'
    """

    try:
        dot = path.rindex('.')
    except ValueError:
        raise ValueError("Error loading object '%s': not a full path" % path)

    module, name = path[:dot], path[dot + 1:]
    mod = import_module(module)

    try:
        obj = getattr(mod, name)
    except AttributeError:
        raise NameError("Module '%s' doesn't define any object named '%s'" % (module, name))

    return obj


if __name__ == '__main__':
    obj = load_object('geeker.commands.abc')
    # obj()



# -*- coding: utf-8 -*-

import click
from geeker import __Version__, __UpdateTime__, __all__, __Description__


@click.command(name='v')
def version():
    """
    Current version.
    """

    click.echo(__Version__)


@click.command(name='t')
def time_stamp():
    """
    The last update time.
    """

    click.echo(__UpdateTime__)


@click.command(name='a')
def about():
    """
    About problems .
    """
    click.echo(__Description__)


@click.command(name='m')
@click.option("--module", prompt="input module name")  # prompt功能:直接弹出一行，让用户输入
def how_use(module):
    """
    How to use some module.
    """
    tmp = 'https://github.com/4379711/functools_lyl#' + str(module).lower().replace(" ", '-')
    click.echo('You can find answer in this url.')
    click.echo(tmp)


@click.command(name='M')
def list_module():
    """
    ALL modules .
    """
    tmp = (i for i in __all__ if not i.startswith('__'))
    for module in tmp:
        click.echo(module)


@click.group()
def base_command():
    pass


# 添加到组
base_command.add_command(how_use)
base_command.add_command(list_module)
base_command.add_command(version)
base_command.add_command(time_stamp)
base_command.add_command(about)

if __name__ == '__main__':
    base_command()



# -*- coding: utf-8 -*-


class Meta(type):
    """
    单例模式的metaclass类
    """

    def __call__(cls, *args, **kwargs):
        if not hasattr(cls, '__instance__'):
            cls.__instance__ = super().__call__(*args, **kwargs)
        return cls.__instance__


class Singleton(metaclass=Meta):
    """这种写法，实例属性是首次创建那个实例的属性"""
    pass


class SingletonOverride:
    """
    单例模式 原理：“类变量对所有对象唯一”
    这种写法，后面的实例属性会覆盖前者的实例属性
    """

    def __new__(cls, *args, **kwargs):
        if not hasattr(cls, '__instance__'):
            cls.__instance__ = super().__new__(cls)
        return cls.__instance__
